﻿#include "main.h"

int WINAPI WinMain(HINSTANCE hInstance,
    HINSTANCE /*hPrevInstance*/,
    LPSTR /*lpCmdLine*/,
    int nCmdShow)
{
    WNDCLASSEXW wc;
    ZeroMemory(&wc, sizeof(wc));
    wc.cbSize = sizeof(WNDCLASSEX);
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = WindowProc;
    wc.hInstance = hInstance;
    wc.hCursor = LoadCursor(nullptr, IDC_ARROW);
    wc.lpszClassName = L"WindowClass";
    RegisterClassExW(&wc);

    RECT wr = { 0, 0, CLIENT_WIDTH, CLIENT_HEIGHT };
    AdjustWindowRect(&wr, WS_OVERLAPPEDWINDOW, FALSE);

    HWND hWnd;
    hWnd = CreateWindowExW(NULL,
        L"WindowClass",
        L"DirectX 11 Test",
        WS_OVERLAPPEDWINDOW,
        300,
        300,
        wr.right - wr.left,
        wr.bottom - wr.top,
        nullptr,
        nullptr,
        hInstance,
        nullptr);

    ShowWindow(hWnd, nCmdShow);
    InitD3D(hWnd);

    MSG msg;
    while (TRUE)
    {
        if (PeekMessage(&msg, nullptr, 0, 0, PM_REMOVE))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);

            if (msg.message == WM_QUIT)
                break;
        }

        // 실제로 프레임 갱신이 일어나는 부분
        RenderFrame();
    }

    ClearD3D();

    return msg.wParam;
}

LRESULT CALLBACK WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        case WM_DESTROY:
        {
            PostQuitMessage(0);
            return 0;
        }
    }

    return DefWindowProc(hWnd, message, wParam, lParam);
}

bool InitD3D(HWND hWnd)
{
    /* 1. D3D11CreateDevice 함수를 이용해서 장치, 즉 ID3D11Device 인터페이스와
    장치 문맥, 즉 ID3D11DeviceContext 인터페이스를 생성한다. */
    if (!InitD3DDevice()) { return false; }

    /* 2. ID3D11Device::CheckMultisampleQualityLevels 메소드를 이용해서
    4X MSAA 품질 수준 지원 여부를 점검한다. */
    if (!Is4XMSAACompatible()) { return false; }

    /* 3. 생성할 Swap Chain의 특성을 서술하는 DXGI_SWAP_CHAIN_DESC 구조체를 채운다. */
    InitSwapChainDesc(hWnd);

    /* 4. 장치를 생성하는데 사용했던 IDXGIFactory 인터페이스를 질의해서
    IDXGISwapChain 인스턴스를 생성한다. */
    if (!InitSwapChain()) { return false; }

    // 5~7번은 후면 버퍼를 파이프라인의 출력 병합기 단계에 묶기 위한 준비작업.
    // 묶어야 Direct3D가 버퍼에 뭔가를 그릴 수 있다.
    /* 5. Swap Chain의 후면 버퍼에 대한 렌더 대상 뷰를 생성한다. */
    InitRenderView();

    /* 6. 깊이&스텐실 버퍼와 그에 연결되는 깊이&스텐실 뷰를 생성한다. */
    // 문을 페인트칠할 때 문고리 부분은 피하기 위해 테이프 등을 붙여놓는데,
    // 이런 역할을 하는 것이 스텐실 버퍼이다. 깊이 버퍼는 말 그대로임.
    // http://wecansoar.tistory.com/103 참고
    if (!InitDepthStencilView()) { return false; }

    /* 7. 렌더 대상 뷰와 깊이&스텐실 뷰를 Direct3D가 사용할 수 있도록
    렌더할 파이프라인의 출력 병합기 단계에 묶는다. */
    BindViewsToOutputMerger();

    /* 8. 뷰포트를 설정한다. */
    InitViewPort();

    /* 9. 파이프라인에 관련된 요소(픽셀&버텍스 셰이더, 입력 레이아웃 등)들을 초기화한다. */
    if (!InitPipeline()) { return false; }

    /* 10. 렌더할 물체들을 초기화한다. */
    InitResources();

    return true;
}
bool InitD3DDevice()
{
    auto vars = DX11StaticVariables::GetInstance();
#if defined(DEBUG) || defined(_DEBUG)
    vars->createDeviceFlags |= D3D11_CREATE_DEVICE_DEBUG;
#endif

    HRESULT hResult = D3D11CreateDevice(
        nullptr,
        D3D_DRIVER_TYPE_HARDWARE,
        nullptr,
        vars->createDeviceFlags,
        nullptr,
        0,
        D3D11_SDK_VERSION,
        &vars->d3dDevice,
        &vars->featureLevel,
        &vars->d3dImmediateContext);

    if (!D3DInitialized(hResult))
    {
        // D3DDevice is not initialized
        return false;
    }

    return true;
}
bool D3DInitialized(HRESULT hResult)
{
    if (FAILED(hResult))
    {
        MessageBoxW(nullptr, L"D3D11CreateDevice Failed", nullptr, 0);
        return false;
    }

    if (DX11StaticVariables::GetInstance()->featureLevel != D3D_FEATURE_LEVEL_11_0)
    {
        MessageBoxW(nullptr, L"DirectX 11 unsupported", nullptr, 0);
        return false;
    }

    return true;
}
// 모든 Direct3D 대응 장치는 4X MSAA를 지원한다.
// 하지만 지원하는 품질 수준은 다를 수 있다.
bool Is4XMSAACompatible()
{
    auto vars = DX11StaticVariables::GetInstance();
    
    HRESULT hResult;
    hResult = vars->d3dDevice->CheckMultisampleQualityLevels(
        DXGI_FORMAT_R8G8B8A8_UNORM, 4, &vars->lMSAAQuality);
    if (FAILED(hResult))
    {
        MessageBoxW(nullptr, L"4XMSAA unsupported", nullptr, 0);
        return false;
    }

    // 지원한다면 lMSAAQuality는 항상 0보다 커야한다.
    assert(vars->lMSAAQuality > 0);

    return true;
}
void InitSwapChainDesc(HWND hWnd)
{
    auto vars = DX11StaticVariables::GetInstance();
    auto &sc = vars->swapChainDesc;
    sc.BufferDesc.Width = vars->clientWidth;
    sc.BufferDesc.Height = vars->clientHeight;
    sc.BufferDesc.RefreshRate.Numerator = 60;
    sc.BufferDesc.RefreshRate.Denominator = 1;
    sc.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
    sc.BufferDesc.ScanlineOrdering = DXGI_MODE_SCANLINE_ORDER_UNSPECIFIED;
    sc.BufferDesc.Scaling = DXGI_MODE_SCALING_UNSPECIFIED;

    if (vars->enableMSAA)
    {
        sc.SampleDesc.Count = 4;
        // 왜 1을 빼야하는진 모름..
        sc.SampleDesc.Quality = vars->lMSAAQuality - 1;
    }
    else
    {
        // MSAA를 적용하지 않음.
        sc.SampleDesc.Count = 1;
        sc.SampleDesc.Quality = 0;
    }

    // 버퍼의 용도가 후면 버퍼 렌더링임을 명시.
    sc.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
    sc.BufferCount = 1; // 1: 이중 버퍼링, 2: 삼중 버퍼링

    sc.OutputWindow = hWnd;
    sc.Windowed = true; // false는 전체화면 모드, true는 창 모드

    // 여러 가지 Swap 방식이 있으니 참고바람
    sc.SwapEffect = DXGI_SWAP_EFFECT_DISCARD;
    sc.Flags = 0;
}
bool InitSwapChain()
{
    // IDXGIFactory의 인스턴스는 dxgi.lib를 링크하고 CreateDXGIFactory()를
    // 써도 얻을 수 있지만 우리가 생성한 장치에 쓰인 인스턴스를 사용해야 한다.
    // 그러지 않으면 오류를 만남 ^^

    auto vars = DX11StaticVariables::GetInstance();
    HRESULT hResult;

    IDXGIDevice * dxgiDevice = nullptr;
    hResult = vars->d3dDevice->QueryInterface(
        __uuidof(IDXGIDevice), reinterpret_cast<void**>(&dxgiDevice));
    if (FAILED(hResult)) { return false; }

    IDXGIAdapter * dxgiAdapter = nullptr;
    hResult = dxgiDevice->GetParent(
        __uuidof(IDXGIAdapter), reinterpret_cast<void**>(&dxgiAdapter));
    if (FAILED(hResult)) { return false; }

    // 지금까지 한 작업은 IDXGIFactory를 얻기 위함이었다!
    IDXGIFactory * dxgiFactory = nullptr;
    hResult = dxgiAdapter->GetParent(
        __uuidof(IDXGIFactory), reinterpret_cast<void**>(&dxgiFactory));
    if (FAILED(hResult)) { return false; }

    hResult = dxgiFactory->CreateSwapChain(
        vars->d3dDevice, &vars->swapChainDesc, &vars->swapChain);
    if (FAILED(hResult)) { return false; }

    // Swap Chain을 생성했으므로 여기서 생성한 COM 인터페이스는 해제해야 됨
    dxgiFactory->Release();
    dxgiAdapter->Release();
    dxgiDevice->Release();

    return true;
}
void InitRenderView()
{
    auto vars = DX11StaticVariables::GetInstance();
    ID3D11Texture2D * backBuffer = nullptr;

    // 이중 버퍼가 아닌 삼중 이상의 버퍼부터는 첫 번째 인자로 대상 버퍼를 고를 수 있다.
    // 이중 버퍼는 전면<->후면 버퍼 구성이라 후면 버퍼가 하나 뿐이므로 0번만 있는 것임.
    vars->swapChain->GetBuffer(
        0, __uuidof(ID3D11Texture2D), reinterpret_cast<void**>(&backBuffer));
    
    // 가운데 인자는 렌더 대상 뷰에 대한 형식을 알려주는 역할을 하는데,
    // 앞의 함수에서 이미 형식을 지정하여 후면버퍼를 생성했으므로 nullptr을 넣은 것임.
    vars->d3dDevice->CreateRenderTargetView(
        backBuffer, nullptr, &vars->renderTargetView);
    
    // 후면 버퍼에 대한 COM 참조 횟수를 줄이기 위해 해제함.
    backBuffer->Release();
}
bool InitDepthStencilView()
{
    auto vars = DX11StaticVariables::GetInstance();
    auto &ds = vars->depthStencilDesc;

    ds.Width = vars->clientWidth;
    ds.Height = vars->clientHeight;
    ds.MipLevels = 1;
    ds.ArraySize = 1;
    ds.Format = DXGI_FORMAT_D24_UNORM_S8_UINT;

    if (vars->enableMSAA)
    {
        // 4XMSAA를 사용할 경우, 반드시 앞에서의 Swap Chain의 설정과 일치해야함

        ds.SampleDesc.Count = 4;
        // 왜 1을 빼는지 모르겠음..
        ds.SampleDesc.Quality = vars->lMSAAQuality - 1;
    }
    else
    {
        // MSAA 사용 안함
        ds.SampleDesc.Count = 1;
        ds.SampleDesc.Quality = 0;
    }

    ds.Usage = D3D11_USAGE_DEFAULT;
    ds.BindFlags = D3D11_BIND_DEPTH_STENCIL;
    
    /*
     CPU가 자원에 자료를 "기록해야" 한다면 D3D11_CPU_ACCESS_WRITE를 지정해야 하며,
    대신 Usage는 D3D11_USAGE_DYNAMIC이나 D3D11_USAGE_STAGING이어야 한다.
     CPU가 자원을 "읽어야" 한다면 D3D11_CPU_ACCESS_READ를 지정해야 하며,
    Usage는 D3D11_USAGE_STAGING이어야 한다. 하지만 깊이&스텐실 버퍼는 GPU만 사용하므로
    0이어도 된다.
     하지만 DYNAMIC과 STAGING은 CPU와 GPU사이에서 메모리가 오가기 때문에
    성능상의 피해가 있으므로 피해야 한다.
    */
    ds.CPUAccessFlags = 0;
    ds.MiscFlags = 0;

    HRESULT hResult;
    hResult = vars->d3dDevice->CreateTexture2D(
        &vars->depthStencilDesc,
        nullptr,
        &vars->depthStencilBuffer);
    if (FAILED(hResult)) { return false; }

    // 렌더뷰에서와 마찬가지로 가운데 인자는 우리가 형식을 정해줬으므로 채울 필요 없음
    hResult = vars->d3dDevice->CreateDepthStencilView(
        vars->depthStencilBuffer,
        nullptr,
        &vars->depthStencilView);
    if (FAILED(hResult)) { return false; }

    return true;
}
void BindViewsToOutputMerger()
{
    auto vars = DX11StaticVariables::GetInstance();
    vars->d3dImmediateContext->OMSetRenderTargets(
        1,
        &vars->renderTargetView,
        vars->depthStencilView);
}
void InitViewPort()
{
    auto vars = DX11StaticVariables::GetInstance();
    auto &vp = vars->viewport;

    vp.TopLeftX = 0.0f;
    vp.TopLeftY = 0.0f;
    vp.Width = static_cast<float>(vars->clientWidth);
    vp.Height = static_cast<float>(vars->clientHeight);
    vp.MinDepth = 0.0f;
    vp.MaxDepth = 1.0f;

    vars->d3dImmediateContext->RSSetViewports(1, &vp);
}
bool InitPipeline()
{
    auto vars = DX11StaticVariables::GetInstance();
    HRESULT result;
    ID3DBlob *VS, *PS;

    result = D3DX11CompileFromFileW(L"shaders.shader", nullptr, nullptr, "VShader", "vs_4_0", 0, 0, nullptr, &VS, nullptr, nullptr);
    if (result != S_OK) { return false; }

    result = D3DX11CompileFromFileW(L"shaders.shader", nullptr, nullptr, "PShader", "ps_4_0", 0, 0, nullptr, &PS, nullptr, nullptr);
    if (result != S_OK) { return false; }

    vars->d3dDevice->CreateVertexShader(VS->GetBufferPointer(), VS->GetBufferSize(), nullptr, &vars->vertexShader);
    vars->d3dDevice->CreatePixelShader(PS->GetBufferPointer(), PS->GetBufferSize(), nullptr, &vars->pixelShader);

    vars->d3dImmediateContext->VSSetShader(vars->vertexShader, nullptr, 0);
    vars->d3dImmediateContext->PSSetShader(vars->pixelShader, nullptr, 0);

    InitInputLayout(VS);

    InitConstantBuffer();

    return true;
}
void InitInputLayout(ID3DBlob* VS)
{
    auto vars = DX11StaticVariables::GetInstance();
    D3D11_INPUT_ELEMENT_DESC ied[] =
    {
        { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
        { "NORMAL", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0 },
    };

    vars->d3dDevice->CreateInputLayout(ied, 2, VS->GetBufferPointer(), VS->GetBufferSize(), &vars->InputLayout);
    vars->d3dImmediateContext->IASetInputLayout(vars->InputLayout);
}
void InitConstantBuffer()
{
    auto vars = DX11StaticVariables::GetInstance();

    D3D11_BUFFER_DESC bd;
    RtlZeroMemory(&bd, sizeof(bd));
    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.ByteWidth = 176; // 상수 버퍼 크기가 아닐까...
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;

    vars->d3dDevice->CreateBuffer(&bd, nullptr, &vars->constantBuffer);
    vars->d3dImmediateContext->VSSetConstantBuffers(0, 1, &vars->constantBuffer);
}
void InitResources()
{
    auto vars = DX11StaticVariables::GetInstance();

    VERTEX OurVertices[] =
    {
        { -1.0f, -1.0f, 1.0f, D3DXVECTOR3(0.0f, 0.0f, 1.0f) }, // side 1
        { 1.0f, -1.0f, 1.0f, D3DXVECTOR3(0.0f, 0.0f, 1.0f) },
        { -1.0f, 1.0f, 1.0f, D3DXVECTOR3(0.0f, 0.0f, 1.0f) },
        { 1.0f, 1.0f, 1.0f, D3DXVECTOR3(0.0f, 0.0f, 1.0f) },

        { -1.0f, -1.0f, -1.0f, D3DXVECTOR3(0.0f, 0.0f, -1.0f) }, // side 2
        { -1.0f, 1.0f, -1.0f, D3DXVECTOR3(0.0f, 0.0f, -1.0f) },
        { 1.0f, -1.0f, -1.0f, D3DXVECTOR3(0.0f, 0.0f, -1.0f) },
        { 1.0f, 1.0f, -1.0f, D3DXVECTOR3(0.0f, 0.0f, -1.0f) },

        { -1.0f, 1.0f, -1.0f, D3DXVECTOR3(0.0f, 1.0f, 0.0f) }, // side 3
        { -1.0f, 1.0f, 1.0f, D3DXVECTOR3(0.0f, 1.0f, 0.0f) },
        { 1.0f, 1.0f, -1.0f, D3DXVECTOR3(0.0f, 1.0f, 0.0f) },
        { 1.0f, 1.0f, 1.0f, D3DXVECTOR3(0.0f, 1.0f, 0.0f) },

        { -1.0f, -1.0f, -1.0f, D3DXVECTOR3(0.0f, -1.0f, 0.0f) }, // side 4
        { 1.0f, -1.0f, -1.0f, D3DXVECTOR3(0.0f, -1.0f, 0.0f) },
        { -1.0f, -1.0f, 1.0f, D3DXVECTOR3(0.0f, -1.0f, 0.0f) },
        { 1.0f, -1.0f, 1.0f, D3DXVECTOR3(0.0f, -1.0f, 0.0f) },

        { 1.0f, -1.0f, -1.0f, D3DXVECTOR3(1.0f, 0.0f, 0.0f) }, // side 5
        { 1.0f, 1.0f, -1.0f, D3DXVECTOR3(1.0f, 0.0f, 0.0f) },
        { 1.0f, -1.0f, 1.0f, D3DXVECTOR3(1.0f, 0.0f, 0.0f) },
        { 1.0f, 1.0f, 1.0f, D3DXVECTOR3(1.0f, 0.0f, 0.0f) },

        { -1.0f, -1.0f, -1.0f, D3DXVECTOR3(-1.0f, 0.0f, 0.0f) }, // side 6
        { -1.0f, -1.0f, 1.0f, D3DXVECTOR3(-1.0f, 0.0f, 0.0f) },
        { -1.0f, 1.0f, -1.0f, D3DXVECTOR3(-1.0f, 0.0f, 0.0f) },
        { -1.0f, 1.0f, 1.0f, D3DXVECTOR3(-1.0f, 0.0f, 0.0f) },
    };

    // create the vertex buffer
    D3D11_BUFFER_DESC bd;
    ZeroMemory(&bd, sizeof(bd));

    bd.Usage = D3D11_USAGE_DYNAMIC;
    bd.ByteWidth = sizeof(VERTEX) * 24;
    bd.BindFlags = D3D11_BIND_VERTEX_BUFFER;
    bd.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    vars->d3dDevice->CreateBuffer(&bd, nullptr, &vars->vertexBuffer);

    // copy the vertices into the buffer
    D3D11_MAPPED_SUBRESOURCE ms;
    vars->d3dImmediateContext->Map(vars->vertexBuffer, NULL, D3D11_MAP_WRITE_DISCARD, NULL, &ms); // map the buffer
    memcpy(ms.pData, OurVertices, sizeof(OurVertices)); // copy the data
    vars->d3dImmediateContext->Unmap(vars->vertexBuffer, NULL);


    // create the index buffer out of DWORDs
    DWORD OurIndices[] =
    {
        0, 1, 2, // side 1
        2, 1, 3,
        4, 5, 6, // side 2
        6, 5, 7,
        8, 9, 10, // side 3
        10, 9, 11,
        12, 13, 14, // side 4
        14, 13, 15,
        16, 17, 18, // side 5
        18, 17, 19,
        20, 21, 22, // side 6
        22, 21, 23,
    };

    // create the index buffer
    bd.Usage = D3D11_USAGE_DYNAMIC;
    bd.ByteWidth = sizeof(DWORD) * 36;
    bd.BindFlags = D3D11_BIND_INDEX_BUFFER;
    bd.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
    bd.MiscFlags = 0;

    vars->d3dDevice->CreateBuffer(&bd, nullptr, &vars->indexBuffer);

    vars->d3dImmediateContext->Map(vars->indexBuffer, NULL, D3D11_MAP_WRITE_DISCARD, NULL, &ms); // map the buffer
    memcpy(ms.pData, OurIndices, sizeof(OurIndices)); // copy the data
    vars->d3dImmediateContext->Unmap(vars->indexBuffer, NULL);
}

void RenderFrame()
{
    auto vars = DX11StaticVariables::GetInstance();

    /* 주변 환경 준비작업 */
    UpdateEnvironment();

    /* 버퍼 초기화 */
    vars->d3dImmediateContext->ClearRenderTargetView(vars->renderTargetView, D3DXCOLOR(0.0f, 0.2f, 0.4f, 1.0f));    // 파란색 배경
    vars->d3dImmediateContext->ClearDepthStencilView(vars->depthStencilView, D3D11_CLEAR_DEPTH, 1.0f, 0);

    /* 표시할 버텍스 버퍼를 선택 */
    UINT stride = sizeof(VERTEX);
    UINT offset = 0;
    vars->d3dImmediateContext->IASetVertexBuffers(0, 1, &vars->vertexBuffer, &stride, &offset);
    vars->d3dImmediateContext->IASetIndexBuffer(vars->indexBuffer, DXGI_FORMAT_R32_UINT, 0);

    /* 사용할 기본도형 위상구조를 선택 */
    vars->d3dImmediateContext->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);

    /* 리소스(도형 등)를 그림 */
    vars->d3dImmediateContext->UpdateSubresource(vars->constantBuffer, 0, nullptr, &vars->envBuffer, 0, 0);
    vars->d3dImmediateContext->DrawIndexed(36, 0, 0);

    /* 스왑체인으로 버퍼를 교환 */
    vars->swapChain->Present(0, 0);
}
void UpdateEnvironment()
{
    auto vars = DX11StaticVariables::GetInstance();
    auto &env = vars->envBuffer;

    env.LightVector = D3DXVECTOR4(0.5f, 0.5f, 0.5f, 1.0f); // 빛의 방향
    env.LightColor = D3DXCOLOR(1.0f, 1.0f, 1.0f, 0.0f); // 빛의 색
    env.AmbientColor = D3DXCOLOR(0.0f, 0.0f, 0.0f, 1.0f); // 주변 빛의 색과 밝기

    D3DXMATRIX matRotate, matView, matProjection;

    static float Time = 0.0f;
    Time += 0.0003f;

    // create a world matrices

    D3DXMatrixRotationY(&matRotate, Time);

    auto camPos = D3DXVECTOR3(0.0f, 3.0f, 5.0f);
    auto eyePos = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
    auto upPos = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
    // create a view matrix

    D3DXMatrixLookAtLH(&matView,
                       &camPos, // the camera position
                       &eyePos, // the look-at position
                       &upPos); // the up direction


    // create a projection matrix

    D3DXMatrixPerspectiveFovLH(&matProjection,
                               static_cast<FLOAT>(D3DXToRadian(45)), // field of view
                               static_cast<FLOAT>(vars->clientWidth) / static_cast<FLOAT>(vars->clientHeight), // aspect ratio
                               1.0f, // near view-plane
                               100.0f); // far view-plane


    // load the matrices into the constant buffer

    env.Final = matRotate * matView * matProjection;
    env.Rotation = matRotate;
}

void ClearD3D()
{
    auto vars = DX11StaticVariables::GetInstance();

    vars->indexBuffer->Release();
    vars->vertexBuffer->Release();
    vars->constantBuffer->Release();
    vars->InputLayout->Release();
    vars->pixelShader->Release();
    vars->vertexShader->Release();
    vars->depthStencilView->Release();
    vars->depthStencilBuffer->Release();
    vars->renderTargetView->Release();
    vars->swapChain->Release();
    vars->d3dImmediateContext->Release();
    vars->d3dDevice->Release();
}